
#if defined(INTERFACE_LIB1_A)
#  error "INTERFACE_LIB1_A defined"
#endif
#if defined(LIB1_A)
#  error "LIB1_A defined"
#endif
#if defined(CONSUMER)
#  error "CONSUMER defined"
#endif

void f1(void)
{
}
