extern void order_E_poison(void);

void order_A(void)
{
  order_E_poison();
}

void order_B(void)
{
  order_E_poison();
}

void order_C(void)
{
  order_E_poison();
}

void order_D(void)
{
  order_E_poison();
}
