/* Distributed under the OSI-approved BSD 3-Clause License.  See accompanying
   file LICENSE.rst or https://cmake.org/licensing for details.  */
#pragma once

enum class cmWIXInstallScope
{
  NONE,
  PER_MACHINE,
  PER_USER
};
