#define C_STD_99 199901L
#define C_STD_11 201112L
#define C_STD_17 201710L
#define C_STD_23 202311L

#ifdef __STDC_VERSION__
#  define C_STD __STDC_VERSION__
#elif defined(__POCC_STDC_VERSION__)
#  define C_STD __POCC_STDC_VERSION__
#endif
